#include "MainWindow.h"
#include <iomanip>
#include <sstream>
#include <QApplication>
#include <QErrorMessage>
#include <QFile>
#include <QSaveFile>
#include <QStringList>
#include <ApplicationServices/ApplicationServices.h>

using namespace std;

namespace {

    /**
     * Return the amount of time in seconds that has elapsed since the
     * last system-wide input event.
     *
     * @return system idle time in seconds
     */
    double
    systemIdleTime() {
        return CGEventSourceSecondsSinceLastEventType(
            kCGEventSourceStateHIDSystemState,
            kCGAnyInputEventType);
    }

}

namespace Harbinger {

    MainWindow::MainWindow(QWidget* parent)
        : QMainWindow(parent),
          m_lastSystemIdleTime(0.0),
          m_systemIdleTimer(this)
    {
        /* Set up the UI. */
        this->setupUi(this);

        /* Set default values for user preferences. */
        this->m_preferences.advanceNotice = 270;  /* 4.5 minutes */
        this->m_preferences.screensaver = 300;    /* 5.0 minutes */
        this->m_preferences.message = "Screensaver Apocalypse";

        /* If possible, override the default values with the
         * preferences saved to disk. */
        this->loadPreferences();

        /* Poll the system idle time very 0.25 seconds. */
        this->connect(&m_systemIdleTimer,
                      SIGNAL(timeout()),
                      this,
                      SLOT(on_pollSystemIdleTime()));
        m_systemIdleTimer.start(250);
    }

    MainWindow::~MainWindow()
    {
        /* m_advanceNoticeDialogs */
        this->clearAdvanceNoticeDialogs();
    }

    void
    MainWindow::on_pollSystemIdleTime()
    {
        double currSystemIdleTime = ::systemIdleTime();

        /*
         * NOTE:
         * ====
         *
         * Showing and hidding the AdvanceNoticeDialog is edge
         * triggered.
         */

        /* Check for the forward transition across the advance notice
         * threshold in which case we should show the dialog. */
        if ((m_lastSystemIdleTime < m_preferences.advanceNotice) &&
            (currSystemIdleTime >= m_preferences.advanceNotice))
        {
            /* Destroy and recreate the dialogs so they are guaranteed
             * to pop up on the current display. */
            clearAdvanceNoticeDialogs();
            createAdvanceNoticeDialogs(currSystemIdleTime);

        /* Check for the reverse transition across the advance notice
         * threshold in which case we should hide the dialog. */
        } else if ((m_lastSystemIdleTime >= m_preferences.advanceNotice) &&
                   (m_lastSystemIdleTime < m_preferences.screensaver) &&
                   (currSystemIdleTime < m_preferences.advanceNotice))
        {
            clearAdvanceNoticeDialogs();

        /* Check for the forward transition across the screensaver
         * enabled threshold in which case we should hide the
         * dialog (because the screensaver just kicked on). */
        } else if ((m_lastSystemIdleTime < m_preferences.screensaver) &&
            (currSystemIdleTime >= m_preferences.screensaver))
        {
            clearAdvanceNoticeDialogs();

        /* Update the countdown timers for each dialog. */
        } else {
            updateMessagesForAdvanceNoticeDialogs(currSystemIdleTime);
        }

        /* Remember the system idle time. */
        m_lastSystemIdleTime = currSystemIdleTime;
    }

    void
    MainWindow::on_actionAboutHarbinger_triggered(bool checked)
    {
        Q_UNUSED(checked);
        m_aboutHarbingerDialog.exec();
    }

    void
    MainWindow::on_actionPreferences_triggered(bool checked)
    {
        Q_UNUSED(checked);

        try {
            /* Display the modal dialog.  To do this, we give the
             * dialog a copy of our canonical set of preferences which
             * the dialog uses to initialize all its widgets.  On
             * success, we copy the changes into our canonical set of
             * preferences and persist them to disk. */
            m_preferencesDialog.setPreferences(m_preferences);
            m_preferencesDialog.exec();
            if (m_preferencesDialog.result() == QDialog::Accepted) {
                Preferences tmp = m_preferencesDialog.preferences();
                if (tmp.advanceNotice >= tmp.screensaver) {
                    throw runtime_error("Advance notice threshold must be less "
                                        "than the screensaver threshold.");
                }
                m_preferences = tmp;
                this->savePreferences();
            }
        } catch (const exception& e) {
            /* Error */
            QErrorMessage* dialog = QErrorMessage::qtHandler();
            dialog->showMessage(QString::fromLatin1(e.what()));
        }
    }

    void
    MainWindow::loadPreferences()
    {
        try {
            /* Open the configuration file. */
            QFile istrm(this->configFileName());
            if (istrm.open(QIODevice::ReadOnly | QIODevice::Text)) {
                /* Deserialize changes from disk. */
                m_preferences = Preferences::deserialize(istrm);
            }
        } catch (const exception& e) {
            /* Error */
            QErrorMessage* dialog = QErrorMessage::qtHandler();
            dialog->showMessage(QString::fromLatin1(e.what()));
        }
    }

    void
    MainWindow::savePreferences() const
    {
        try {

            /* Try to create the configuration directory if it doesn't
             * already exist. */
            QDir configDir(this->configDirName());
            configDir.mkpath(QString::fromLatin1("."));

            /* Open the configuration file. */
            QString fname = this->configFileName();
            QSaveFile ostrm(fname);
            if (!ostrm.open(QIODevice::WriteOnly | QIODevice::Text)) {
                ostringstream errmsg;
                errmsg << "Unable to open configuration file: "
                       << qPrintable(fname);
                throw runtime_error(errmsg.str());
            }

            /* Write changes to disk. */
            m_preferences.serialize(ostrm);

            /* Commit changes. */
            if (!ostrm.commit()) {
                ostringstream errmsg;
                errmsg << "Unable to save configuration to file: "
                       << qPrintable(fname);
                throw runtime_error(errmsg.str());
            }

        } catch (const exception& e) {
            /* Error */
            QErrorMessage* dialog = QErrorMessage::qtHandler();
            dialog->showMessage(QString::fromLatin1(e.what()));
        }
    }

    QString
    MainWindow::configDirName() const
    {
        QStringList pathElements;
        pathElements << QDir::homePath();
        pathElements << QString::fromLatin1(".config");
        pathElements << QString::fromLatin1("Harbinger");
        return pathElements.join(QDir::separator());
    }

    QString
    MainWindow::configFileName() const
    {
        return this->configDirName() +
               QDir::separator() +
               QString::fromLatin1("config.xml");
    }

    void
    MainWindow::createAdvanceNoticeDialogs(double currSystemIdleTime)
    {
        typedef QList<QScreen*> ScreenList;
        ScreenList::const_iterator iter;

        /* Iterate over each screen. */
        ScreenList screens = QApplication::screens();
        for (iter = screens.begin() ; iter != screens.end() ; ++iter) {
            /* Add a new AdvanceNoticeDialog to the list. */
            m_advanceNoticeDialogs.push_back(new AdvanceNoticeDialog());
            /* Create an aliases for the screen and the dialog that
             * will be displayed on that screen. */
            QScreen* screen = *iter;
            AdvanceNoticeDialog* dialog = m_advanceNoticeDialogs.back();
            /* Configure the dialog to display on this screen. */
            QString msg = generateAdvanceNoticeMessage(currSystemIdleTime);
            dialog->messageLabel->setText(msg);
            dialog->setGeometry(screen->geometry());
            dialog->show();
            dialog->raise();
            dialog->activateWindow();
        }
    }

    void
    MainWindow::clearAdvanceNoticeDialogs()
    {
        /* Delete each AdvanceNoticeDialog instance. */
        AdvanceNoticeDialogList::iterator iter;
        for (iter = m_advanceNoticeDialogs.begin();
             iter != m_advanceNoticeDialogs.end();
             ++iter)
        {
            delete *iter;
            *iter = NULL;
        }

        /* Clear the list of AdvanceNoticeDialog instances. */
        m_advanceNoticeDialogs.clear();
    }

    void
    MainWindow::updateMessagesForAdvanceNoticeDialogs(double currSystemIdleTime)
    {
        /* Update the message for each AdvanceNoticeDialog instance. */
        AdvanceNoticeDialogList::iterator iter;
        for (iter = m_advanceNoticeDialogs.begin();
             iter != m_advanceNoticeDialogs.end();
             ++iter)
        {
            AdvanceNoticeDialog* dialog = *iter;
            QString msg = generateAdvanceNoticeMessage(currSystemIdleTime);
            dialog->messageLabel->setText(msg);
        }
    }

    QString
    MainWindow::generateAdvanceNoticeMessage(double currSystemIdleTime)
    {
        ostringstream rv;
        /* Calculate how much time remains before the screensaver starts. */
        unsigned long countdown =
            (unsigned long)(m_preferences.screensaver - currSystemIdleTime);
        unsigned long seconds = countdown % 60;
        unsigned long minutes = countdown / 60 % 60;
        unsigned long hours = countdown / 3600;
        rv << m_preferences.message << "\n"
           << setw(2) << setfill('0') << hours << ":"
           << setw(2) << setfill('0') << minutes << ":"
           << setw(2) << setfill('0') << seconds;
        return QString::fromLatin1(rv.str().c_str());
    }

}
