#ifndef MAIN_WINDOW_H
#define MAIN_WINDOW_H

#include "AboutHarbingerDialog.h"
#include "AdvanceNoticeDialog.h"
#include "Preferences.h"
#include "PreferencesDialog.h"
#include <QList>
#include <QString>
#include <QtCore/QtGlobal>
#include <QTimer>
#include <QtWidgets/QtWidgets>

#include "ui_MainWindow.h"

namespace Harbinger {

    /**
     * Main Window
     */
    class MainWindow : public QMainWindow,
                       public Ui::MainWindow
    {
        Q_OBJECT

    public:

        /**
         * Construct a new <c>MainWindow</c> instance having parent
         * <c>parent</c>.
         *
         * @param[in] parent parent
         */
        MainWindow(QWidget* parent = NULL);

        /**
         * Destructor
         */
        virtual
        ~MainWindow();

    private slots:

        /**
         * Slot used to poll the system idle time so we can detect
         * when the screensaver is about to be enabled.
         */
        void
        on_pollSystemIdleTime();

        /**
         * Slot used to display <c>m_aboutHarbingerDialog</c>.
         *
         * @param[in] checked not used
         */
        void
        on_actionAboutHarbinger_triggered(bool checked = false);

        /**
         * Slot used to display <c>m_preferencesDialog</c>.
         *
         * @param[in] checked not used
         */
        void
        on_actionPreferences_triggered(bool checked = false);

    private:

        /**
         * Load the preferences <c>m_preferences</c> from the
         * configuration file <c>configFileName()</c>.  If the
         * configuration file does not exist, this method silently
         * returns without altering <c>m_preferences</c>
         */
        void
        loadPreferences();

        /**
         * Save the preferences <c>m_preferences</c> to the
         * configuration file <c>configFileName()</c>.
         */
        void
        savePreferences() const;

        /**
         * Return the absolute path to the directory where preferences
         * are saved.
         *
         * @return absolute path to directory where preferences are saved
         */
        QString
        configDirName() const;

        /**
         * Return the absolute path of the file where preferences are
         * saved.
         *
         * @return absolute path of file where preferences are saved
         */
        QString
        configFileName() const;

        /**
         * Create all the <c>AdvanceNoticeDialog</c> instances (one
         * per screen).  The current system idle time
         * <c>currSystemIdleTime</c> is used to dynamically update the
         * countdown timer that is part of the message for each
         * dialog.
         *
         * @param[in] currSystemIdleTime current system idle time
         */
        void
        createAdvanceNoticeDialogs(double currSystemIdleTime);

        /**
         * Destroy all <c>AdvanceNoticeDialog</c> instances on the
         * <c>m_advanceNoticeDialogs</c> list.
         */
        void
        clearAdvanceNoticeDialogs();

        /**
         * Update the messages for each <c>AdvanceNoticeDialog</c>
         * instance on the <c>m_advanceNoticeDialogs</c> list.  The
         * main point of this is to dynamically update the countdown
         * timer that is part of the message for each dialog based on
         * the current system idle time <c>currSystemIdleTime</c>.
         *
         * @param[in] currSystemIdleTime current system idle time
         */
        void
        updateMessagesForAdvanceNoticeDialogs(double currSystemIdleTime);

        /**
         * Return the message used to give the user advance notice of
         * when the screensaver will be engaged.  The main point of
         * this is to dynamically update the countdown timer that is
         * part of the message based on the current system idle time
         * <c>currSystemIdleTime</c>.
         *
         * @param[in] currSystemIdleTime current system idle time
         */
        QString
        generateAdvanceNoticeMessage(double currSystemIdleTime);

        /**
         * Type definition for a list of <c>AdvanceNoticeDialog</c>
         * instances.
         */
        typedef QList<AdvanceNoticeDialog*> AdvanceNoticeDialogList;

        /**
         * Dialog that tells about the Harbinger application.
         */
        AboutHarbingerDialog m_aboutHarbingerDialog;

        /**
         * Dialogs used to give the user warning that the screensaver
         * is about to start.
         */
        AdvanceNoticeDialogList m_advanceNoticeDialogs;

        /**
         * Dialog used to edit preferences.
         */
        PreferencesDialog m_preferencesDialog;

        /**
         * Canonical set of preferences.
         */
        Preferences m_preferences;

        /**
         * The value of the system idle time the last time it was
         * polled.
         */
        double m_lastSystemIdleTime;

        /**
         * Timer used to poll the system idle time.
         */
        QTimer m_systemIdleTimer;

    };

}

#endif  /*  MAIN_WINDOW_H  */
