#ifndef XML_UTIL_H
#define XML_UTIL_H

/*
 * Copyright (c) 2016 Paul Serice
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the author or authors be held liable for
 * any damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *     1. The origin of this software must not be misrepresented; you
 *        must not claim that you wrote the original software. If you use
 *        this software in a product, an acknowledgment in the product
 *        documentation would be appreciated but is not required.
 *
 *     2. Altered source versions must be plainly marked as such and
 *        must not be misrepresented as being the original software.
 *
 *     3. This notice may not be removed or altered from any source
 *        distribution.
 */

#include <cstdlib>
#include <iomanip>
#include <sstream>
#include <stdexcept>
#include <string>
#include <typeinfo>
#include <QFile>
#include <QString>
#include <QXmlStreamReader>

/**
 * The <code>XmlUtil</code> namespace has utility functions that are
 * useful when using <code>QXmlStreamReader</code> to write a
 * recursive descent parse with error checking and line number
 * reporting.  The main features are as follows: <p>
 *
 * <pre>
 *     XmlFileReader(fname)
 *     getXmlFileName()
 *     consumeEmptyElement()
 *     nextToken()
 *     acceptToken()
 *     expectToken()
 *     acceptAttribute()
 *     expectAttribute()
 *     acceptText()
 *     expectText()
 *     expectTextElement()
 *     stringTo<T>()
 * </pre>
 *
 * @note Code in the <code>XmlUtil</code> namespace is not like most
 * Qt code in that it throws exceptions.  Make sure any code that uses
 * the <code>XmlUtil</code> namespace is wrapped in a "try/catch"
 * block.
 */
namespace XmlUtil {

    /*
     * Forward References
     */

    void
    consumeEmptyElement(QXmlStreamReader& xml,
                        const char* tokenName,
                        bool skipPureWhiteSpace,
                        bool skipComments);

    void
    nextToken(QXmlStreamReader& xml,
              bool skipPureWhiteSpace,
              bool skipComments);

    bool
    acceptToken(QXmlStreamReader& xml,
                QXmlStreamReader::TokenType tokenType,
                const char* tokenName);

    void
    expectToken(QXmlStreamReader& xml,
                QXmlStreamReader::TokenType tokenType,
                const char* tokenName);

    bool
    acceptAttribute(QXmlStreamReader& xml,
                    const std::string& attrName,
                    std::string& attrValue);

    void
    expectAttribute(QXmlStreamReader& xml,
                    const std::string& attrName,
                    std::string& attrValue);

    bool
    acceptText(QXmlStreamReader& xml,
               std::string& text);

    std::string
    expectText(QXmlStreamReader& xml);

    std::string
    expectTextElement(QXmlStreamReader& xml,
                      const std::string& tokenName);

    /**
     * This class is a wrapper around <code>QXmlStreamReader</code> to
     * make it more convenient to open a file as an XML stream.
     */
    class XmlFileReader : public QXmlStreamReader
    {
    public:

        /**
         * Create a new default <code>XmlFileReader</code> instance.
         * Use <code>open()</code> to open a file as an XML stream.
         */
        XmlFileReader()
            : f(NULL)
        {
            /* empty */
        }

        /**
         * Create a new <code>XmlFileReader</code> instance that
         * automatically opens the XML file <code>fname</code> as an
         * XML stream.
         *
         * @param[in] fname file name
         */
        XmlFileReader(const char* fname)
            : f(NULL)
        {
            /* Open the file. */
            this->open(fname);
        }

        /**
         * Create a new <code>XmlFileReader</code> instance that
         * automatically opens the XML file <code>fname</code> as an
         * XML stream.
         *
         * @param[in] fname file name
         */
        XmlFileReader(const QString& fname)
            : f(NULL)
        {
            /* Open the file. */
            this->open(fname);
        }

        /**
         * Destructor.  This method cleans up all internally allocated
         * resources including the underlying <code>QFile</code>.
         */
        ~XmlFileReader()
        {
            if (f) {
                delete f;
                f = NULL;
            }
        }

        /**
         * Open the file <code>fname</code> as an XML stream.  If an
         * error occurs, an exception is thrown.
         *
         * @param[in] fname file name
         */
        inline void
        open(const char* fname)
        {
            this->open(QString::fromUtf8(fname));
        }

        /**
         * Open the file <code>fname</code> as an XML stream.  If an
         * error occurs, an exception is thrown.
         *
         * @param[in] fname file name
         */
        inline void
        open(const QString& fname)
        {
            /* Close any previously opened file. */
            this->close();
            /* Create a new QFile object only loosely associated with
             * the target file at this point. */
            f = new QFile(fname);
            if (!f) {
                throw std::runtime_error(
                    "XmlFileReader::open(): out of memory");
            }
            /* Open the target file. */
            f->open(QIODevice::ReadOnly | QIODevice::Text);
            if (f->error() != QFile::NoError) {
                std::ostringstream msg;
                msg << "XmlFileReader::open(): "
                    << "Unable to open file "
                    << "\"" << qPrintable(fname) << "\": "
                    << qPrintable(f->errorString());
                throw std::runtime_error(msg.str());
            }
            /* Attach the file to our QXmlStreamReader. */
            this->setDevice(f);
        }

        /**
         * Close the XML stream detaching it from the underlying file.
         */
        inline void
        close()
        {
            if (f) {
                /* Detach the file from our QXmlStreamReader. */
                this->clear();
                /* Delete the file. */
                delete f;
                f = NULL;
            }
        }

    private:

        /* Hide these. */
        XmlFileReader(const XmlFileReader&);
        XmlFileReader& operator=(const XmlFileReader&);

        /**
         * The underlying file being treated as an XML stream.
         */
        QFile* f;

    };

    /**
     * Return the string representation of the
     * <code>QXmlStreamReader</code> token type
     * <code>tokenType</code>.
     *
     * @param[in] tokenType token type
     * @return string representation of the token type
     */
    inline const char*
    tokenTypeToString(QXmlStreamReader::TokenType tokenType)
    {
        const char* rv = NULL;
        switch (tokenType) {
        case QXmlStreamReader::NoToken:
            rv = "NoToken";
            break;
        case QXmlStreamReader::Invalid:
            rv = "Invalid";
            break;
        case QXmlStreamReader::StartDocument:
            rv = "StartDocument";
            break;
        case QXmlStreamReader::EndDocument:
            rv = "EndDocument";
            break;
        case QXmlStreamReader::StartElement:
            rv = "StartElement";
            break;
        case QXmlStreamReader::EndElement:
            rv = "EndElement";
            break;
        case QXmlStreamReader::Characters:
            rv = "Characters";
            break;
        case QXmlStreamReader::Comment:
            rv = "Comment";
            break;
        case QXmlStreamReader::DTD:
            rv = "DTD";
            break;
        case QXmlStreamReader::EntityReference:
            rv = "EntityReference";
            break;
        case QXmlStreamReader::ProcessingInstruction:
            rv = "ProcessingInstruction";
            break;
        default:
            rv = "<unknown>";
            break;
        }
        return rv;
    }

    /**
     * Return the name of the XML file being parsed.
     *
     * @param[in] xml xml parser
     * @return name of the XML file being parsed
     */
    inline QString
    getXmlFileName(QXmlStreamReader& xml)
    {
        QFile* file = dynamic_cast<QFile*>(xml.device());
        return (file ?
                file->fileName() :
                QString::fromLatin1("<unknown_device>"));
    }

    /**
     * To make it easier to port your code, use this function instead
     * of <code>nextToken()</code> when consuming "empty element"
     * tokens.  This function consumes an "empty element" having the
     * name <code>tokenName</code> from the XML stream
     * <code>xml</code>.  If <code>skipPureWhiteSpace</code> is
     * <code>true</code>, character sections that are pure white space
     * are skipped allowing your parser to ignore the white space that
     * is typically added to improve the readability of the XML file.
     * If <code>skipComments</code> is <code>true</code>, comments are
     * also skipped. <p>
     *
     * If an error occurs, an exception is thrown.  The exception
     * specifies the file name <code>fname</code>, line number, and
     * column where the error occurred. <p>
     *
     * An "empty element" is either one tag having the form
     * <code>&lt;foo/&gt;</code> or equivalently two tags having the
     * form <code>&lt;foo&gt;&lt;/foo&gt;</code>.  Sane API's (like
     * Java's StAX) generate both START_ELEMENT and END_ELEMENT tokens
     * regardless of whether the elements are represented in XML as
     * one tag or two, but some API's (like .NET's XmlReader and
     * libxml2's xmlTextReader) generate just a START_ELEMENT token
     * for the one-tag form and START_ELEMENT and END_ELEMENT tokens
     * for the two-tag form. <p>
     *
     * It would be better to just configure the XML parsers to always
     * generate both token, but this does not always seem possible.
     * Instead, this function effectively always collapses the two
     * events into one.  This makes it easier to port your code to XML
     * parsers that only report one collapsed token. <p>
     *
     * Note that technically parsers that treat the one-tag form
     * differently from the two-tag form only consider the one-tag
     * form to constitute an "empty element" even though both forms
     * are empty. <p>
     *
     * @param[in] xml xml parser
     * @param[in] tokenName expected token name
     * @param[in] skipPureWhiteSpace whether to skip pure white space
     * @param[in] skipComments whether to skip comments
     * @see nextToken()
     */
    inline void
    consumeEmptyElement(QXmlStreamReader& xml,
                        const char* tokenName,
                        bool skipPureWhiteSpace = true,
                        bool skipComments = true)
    {
        /* Make sure the START_ELEMENT is current. */
        expectToken(xml, QXmlStreamReader::StartElement, tokenName);

        /* Read the next token. */
        nextToken(xml, skipPureWhiteSpace, skipComments);

        /* Make sure the END_ELEMENT is current. */
        expectToken(xml, QXmlStreamReader::EndElement, tokenName);

        /* Read the next token. */
        nextToken(xml, skipPureWhiteSpace, skipComments);
    }

    /**
     * Read the next token from the XML stream.  If
     * <code>skipPureWhiteSpace</code> is <code>true</code>, character
     * sections that are pure white space are skipped allowing your
     * parser to ignore the white space that is typically added to
     * improve the readability of the XML file.  If
     * <code>skipComments</code> is <code>true</code>, comments are
     * also skipped.  If an error occurs, an exception is thrown. <p>
     *
     * You MUST NOT try to consume the <code>EndDocument</code> token.
     * Attempting to do so will generate a "Premature end of document"
     * error because there are no more tokens to take its place. <p>
     *
     * @param[in] xml xml parser
     * @param[in] skipPureWhiteSpace whether to skip pure white space
     * @param[in] skipComments whether to skip comments
     */
    inline void
    nextToken(QXmlStreamReader& xml,
              bool skipPureWhiteSpace = true,
              bool skipComments = true)
    {
        /* Try to get the next XML token skipping the tokens that
         * represent just white space if necessary. */
        for (;;) {
            xml.readNext();
            if (skipPureWhiteSpace && xml.isWhitespace()) {
                continue;
            }
            if (skipComments && xml.isComment()) {
                continue;
            }
            break;
        }

        /* Check for errors. */
        if (xml.hasError()) {
            std::ostringstream msg;
            msg << qPrintable(getXmlFileName(xml)) << ":"
                << xml.lineNumber() << ":" << xml.columnNumber() << ": "
                << "Unable to get next XML token: "
                << qPrintable(xml.errorString());
            throw std::runtime_error(msg.str());
        }
    }

    /**
     * Accept the current token from the XML stream <code>xml</code> if
     * its token type matches <code>tokenType</code> and its local name
     * matches <code>tokenName</code> (unless <code>tokenName</code> is
     * <code>NULL</code> in which case the token name is not compared).
     * If the token is accepted, <code>true</code> is returned; otherwise,
     * <code>false</code> is returned.
     *
     * @param[in] xml xml parser
     * @param[in] tokenType acceptable token type
     * @param[in] tokenName acceptable token name
     * @return whether the token was accepted
     */
    inline bool
    acceptToken(QXmlStreamReader& xml,
                QXmlStreamReader::TokenType tokenType,
                const char* tokenName = NULL)
    {
        return (tokenType == xml.tokenType())
            && ((tokenName == NULL) ||
                (QString::fromLatin1(tokenName) == xml.name()));
    }

    /**
     * This function is similar to <code>acceptToken()</code> except
     * it throws an exception if there is no match.  The exception
     * specifies the file name, line number, and column where the
     * match fails and the expected and actual tokens that failed to
     * match.
     *
     * @param[in] xml xml parser
     * @param[in] tokenType expected token type
     * @param[in] tokenName expected token name
     */
    inline void
    expectToken(QXmlStreamReader& xml,
                QXmlStreamReader::TokenType tokenType,
                const char* tokenName = NULL)
    {
        /* Load the next token only if the current token has the
         * expected name; otherwise throw an exception. */
        if (!acceptToken(xml, tokenType, tokenName)) {
            std::ostringstream msg;

            /* Preamble for the error message. */
            msg << qPrintable(getXmlFileName(xml)) << ":"
                << xml.lineNumber() << ":" << xml.columnNumber() << ": "
                << "Expected ";

            /* What was expected. */
            if (tokenName &&
                (tokenType == QXmlStreamReader::StartElement))
            {
                msg << "tag <" << tokenName << "> ";
            } else if (tokenName &&
                       (tokenType == QXmlStreamReader::EndElement))
            {
                msg << "tag </" << tokenName << "> ";
            } else {
                msg << "token having type \""
                    << tokenTypeToString(tokenType)
                    << "\" ";
            }

            /* What was actually read. */
            msg << "but found ";
            if (xml.tokenType() == QXmlStreamReader::StartElement) {
                msg << "tag <" << qPrintable(xml.name().toString()) << ">.";
            } else if ((xml.tokenType() == QXmlStreamReader::EndElement)) {
                msg << "tag </" << qPrintable(xml.name().toString()) << ">.";
            } else {
                msg << "token having type \""
                    << tokenTypeToString(xml.tokenType())
                    << "\" ";
            }

            throw std::runtime_error(msg.str());
        }
    }

    /**
     * Return true if the current token for the xml parser <c>xml</c>
     * is a <c>QXmlStreamReader::StartElement</c> token that has an
     * attribute named <c>attrName</c> and save the value of the
     * attribute in <c>attrValue</c>; otherwise, return <c>false</c>.
     *
     * @param[in] xml xml parser
     * @param[in] attrName acceptable attribute name
     * @param[out] attrValue value of acceptable attribute
     * @return whether the current XML element has the attribute
     */
    inline bool
    acceptAttribute(QXmlStreamReader& xml,
                    const std::string& attrName,
                    std::string& attrValue)
    {
        bool rv = false;

        /* Check if the attribute exists. */
        if (xml.isStartElement()) {

            /* Get the attributes for the start element tag. */
            QXmlStreamAttributes attrs = xml.attributes();

            /* Check if "attrName" is one of the attributes. */
            QString qsAttrName = QString::fromUtf8(attrName.c_str());
            if (attrs.hasAttribute(qsAttrName)) {

                /* Return the attribute value. */
                attrValue = qPrintable(attrs.value(qsAttrName).toString());

                /* Success. */
                rv = true;
            }
        }

        return rv;
    }

    /**
     * Expect the current token for the xml parser <c>xml</c> to be a
     * <c>QXmlStreamReader::StartElement</c> token that has an
     * attribute named <c>attrName</c> and save the value of the
     * attribute in <c>attrValue</c>; otherwise, an exception is
     * thrown.
     *
     * @param[in] xml xml parser
     * @param[in] attrName expected attribute name
     * @param[out] attrValue value of expected attribute
     */
    inline void
    expectAttribute(QXmlStreamReader& xml,
                    const std::string& attrName,
                    std::string& attrValue)
    {
        /* Try to accept the attribute. */
        if (!acceptAttribute(xml, attrName, attrValue)) {
            /* Otherwise, throw an exception. */
            std::ostringstream msg;
            msg << qPrintable(getXmlFileName(xml)) << ":"
                << xml.lineNumber() << ":" << xml.columnNumber() << ": "
                << "Expected attribute "
                << "\"" << attrName << "\" not found." ;
            throw std::runtime_error(msg.str());
        }
    }

    /**
     * Return <c>true</c> if the current token for the xml parser
     * <c>xml</c> is text, and consume and save all consecutive text
     * tokens to <c>text</c>.  Note that unlike the other "accept"
     * functions, this function consumes what it accepts.
     *
     * @param[in] xml xml parser
     * @param[out] text text
     * @return whether text was read from the xml parser
     */
    inline bool
    acceptText(QXmlStreamReader& xml,
               std::string& text)
    {
        bool rv = false;

        /* Clear the text. */
        text.clear();

        /* Consume as much text as possible and save it to "text". */
        while (xml.isCharacters()) {
            text += qPrintable(xml.text().toString());
            nextToken(xml);
            rv = true;
        }

        return rv;
    }

    /**
     * Expect the current token for the xml parser <c>xml</c> to be
     * text, and consume and return all consecutive text tokens;
     * otherwise, an exception is thrown.  Note that unlike the other
     * "expect" functions, this function consumes what it expects.
     *
     * @param[in] xml xml parser
     * @return text read from the xml parser
     * @see expectTextElement()
     */
    inline std::string
    expectText(QXmlStreamReader& xml)
    {
        std::string rv;

        /* Try to accept the text. */
        if (!acceptText(xml, rv)) {
            /* Otherwise, throw an exception. */
            std::ostringstream msg;
            msg << qPrintable(getXmlFileName(xml)) << ":"
                << xml.lineNumber() << ":" << xml.columnNumber() << ": "
                << "Expected text.";
            throw std::runtime_error(msg.str());
        }

        return rv;
    }

    /**
     * Return the text surrounded by the XML element having name
     * <code>tokenName</code>. <p>
     *
     * This function is designed to make it easy to recover the text
     * generated by <code>QXmlStreamWriter::writeTextElement()</code>.
     * For example, <code>writeTextElement("foo", "bar")</code> will
     * generate the "bar" text surrounded by the opening and closing
     * XML tags for "foo".  By calling <code>expectTextElement(xml,
     * "foo")</code>, the "bar" text can be recovered. <p>
     *
     * This function uses <code>expectToken()</code> to verify the
     * opening and closing tags of the surrounding XML.  This function
     * only verifies the closing XML tag without consuming it.  This
     * is similar to the way the rest of the API works in that
     * <code>nextToken()</code> needs to be called to move passed the
     * tag being inspected. <p>
     *
     * @param[in] xml xml parser
     * @param[in] tokenName expected name of XML element surrounding the text
     * @return text surrounded by the expected XML element
     */
    inline std::string
    expectTextElement(QXmlStreamReader& xml,
                      const std::string& tokenName)
    {
        std::string rv;

        /* Expect the start element token. */
        expectToken(xml, QXmlStreamReader::StartElement, tokenName.c_str());
        nextToken(xml);

        /* If we are immediately at the end element, then the text is
         * empty, and there is nothing else to do. */
        if (!acceptToken(xml, QXmlStreamReader::EndElement, tokenName.c_str()))
        {
            /* We are not immediately at the end element so we should
             * expect some text. */
            rv = expectText(xml);

            /* Expect the end element token but do not consume it. */
            expectToken(xml, QXmlStreamReader::EndElement, tokenName.c_str());
        }

        return rv;
    }

    /**
     * Return a copy of the Qt string <code>qs</code> as a C-style
     * string.  The caller is responsible for calling
     * <code>free()</code> to deallocate the array of characters
     * returned.
     *
     * @param[in] qs Qt string
     * @return C-style string
     */
    inline char*
    qstringToCString(const QString& qs)
    {
        char* rv = NULL;

        /* Convert the Qt-style string to a C++-style string.  This
         * intermediate step is necessary because the string returned
         * by qPrintable() is short-lived. */
        std::string cs = qPrintable(qs);

        /* Allocate space for the C-style string and NUL-terminate. */
        rv = (char*)calloc(cs.length() + 1, sizeof(*rv));

        /* Copy the C++-style string to the C-style string. */
        for (size_t i = 0 ; i < cs.length() ; ++i) {
            rv[i] = cs[i];
        }

        return rv;
    }

    /**
     * Convert the Qt string <code>qs</code> to the parameterized
     * number type <code>T</code>.  If an error occurs, an exception
     * is thrown using <code>xml</code> to report the name of the XML
     * file and the line and column for the error. <p>
     *
     * For example, use something like the following to convert a
     * string to a <code>double</code>: <p>
     *
     * <pre>
     *     x = stringTo<double>(xml, qs);
     * </pre>
     *
     * @param[in] xml xml parser
     * @param[in] qs Qt string
     * @return <code>qs</code> convert to the numeric type <code>T</code>
     */
    template <typename T> inline T
    stringTo(QXmlStreamReader& xml,
             QString qs)
    {
        T x;
        std::string s;
        std::istringstream istrm(qPrintable(qs));
        istrm.setf(std::ios_base::boolalpha);
        /* First, try extract a "T" from the string.  If that
         * succeeds, then try not to extract any more characters from
         * the string. */
        if (!(istrm >> x) || (istrm >> s)) {
            std::ostringstream msg;
            msg << qPrintable(getXmlFileName(xml)) << ":"
                << xml.lineNumber() << ":" << xml.columnNumber() << ": "
                << "Unable to convert to type "
                << "\"" << typeid(T).name() << "\" "
                << "from string "
                << "\"" << qPrintable(qs) << "\".";
            throw std::runtime_error(msg.str());
        }
        return x;
    }

}

#endif  /*  XML_UTIL_H  */
